(* 
 * Copyright 2007 Robert R Schneck <schneck@gmail.com>
 * Copyright (c) 2003-2004 Tim-Philipp Mueller <tim at centricular dot net>
 * 
 * This tutorial may be redistributed and modified freely in any form, 
 * as long as all authors are given due credit for their work and all 
 * non-trivial changes by third parties are clearly marked as such either
 * within the document (e.g. in a revision history), or at an external
 * and publicly accessible place that is refered to in the document
 * (e.g. a CVS repository).
 * 
 * This is an OCaml version of the custom model example which is part
 * of the GTK+ 2.0 Tree View Tutorial by Tim-Philipp Mueller (available
 * at http://scentric.net/tutorial/).  The original was 3 C files, here
 * combined into the one OCaml file main.ml.  The original C comments
 * are included here as well.
 * 
 * You need my custom_model.c and custom_model.ml to build, using the commands
 * 
 * gcc -c  -I /opt/local/lib/ocaml `pkg-config --cflags gtk+-2.0` -o custom_model_c.o custom_model.c
 * ocamlc -c -I +lablgtk2 -thread -o custom_model.cmo custom_model.ml
 * ocamlc -c -I +lablgtk2 -thread -o main.cmo main.ml
 * ocamlc -I +lablgtk2 -thread -custom -g unix.cma threads.cma lablgtk.cma gtkThread.cmo custom_model.cmo custom_model_c.o main.cmo
 * 
 * Then you can run a.out and see the example in action.
 *)


(*****************
 * custom-list.h *
 *****************)

(* /* CustomRecord: this structure represents a row */ *)

type custom_record = {
  (* /* data - you can extend this */ *)
  mutable name : string;
  (* mutable name_collate_key : string; *) (* not used *)
  mutable year_born : int;
  (* /* admin stuff used by the custom list model */ *)
  mutable pos : int;  (* /* pos within the array */ *)
}

(* /* The data columns that we export via the tree model interface */ *)

let col_record = 0
let col_name = 1
let col_year_born = 2
let n_columns = 3

(* Note that LablGTK's view_column function expects an argument of type
'a GTree.column.  The column has three fields: index, the numerical
index into the model; conv, a statically declared type of the column 
(which determines 'a); and creator, a stamp which could identify which 
model this GTree.column belongs to.

The backend GTK C interface only cares about the numerical index.
Most of LablGTK only cares about the index too, for instance the code
for view_columns ignores the type information.  LablGTK's model class
does use this information, in the "get" method.  We won't use the
get method, and so we will create fake GTree.columns with irrelevant
conv and creator fields just to have something to pass to GTree.view_column. 

Incidentally the actual work of staying type-sensible happens in 
the coding of custom_get_column_type and custom_get_value. *)

let index_to_mock_column (index:int) : int GTree.column =
  { GTree.index = index; GTree.conv = Gobject.Data.int; GTree.creator = -1 }

(*****************
 * custom_list.c *
 *****************)

class custom_list =
object (self)
  inherit [custom_record,custom_record,unit,unit] Custom_model.custom_model (Custom_model.create ())
  method encode_iter (cr:custom_record) : custom_record * unit * unit = (cr, (), ())
  method decode_iter (cr:custom_record) () () : custom_record = cr

  (* /* number of rows that we have   */ *)
  val mutable num_rows : int = 0
  (* /* a dynamically allocated array of pointers to 
      *   the CustomRecord structure for each row    */ *)
  val mutable rows : custom_record array = [||]

  (* /* These two fields are not absolutely necessary, but they    */
     /*   speed things up a bit in our get_value implementation    */  *)
  val column_types : Gobject.g_type array =
    [| Gobject.Type.caml;
       Gobject.Type.of_fundamental `STRING;
       Gobject.Type.of_fundamental `INT
    |]  

  method custom_get_flags : GtkEnums.tree_model_flags list = [`LIST_ONLY; `ITERS_PERSIST]

  method custom_get_n_columns :int = n_columns

  method custom_get_column_type (n:int) : Gobject.g_type =
    if 0 <= n && n < n_columns then column_types.(n) 
    else Gobject.Type.of_fundamental `INVALID

  method custom_get_iter (path:Gtk.tree_path) : custom_record option =
    (* /* we do not allow children */ *)
    (* /* depth 1 = top level; a list only has top level nodes and no children */ *)
    assert (GTree.Path.get_depth path = 1);
    let n = (GTree.Path.get_indices path).(0) in (* /* the n-th top level row */ *)
    if n >= num_rows || n < 0 then None else begin
      let record = rows.(n) in
      assert (record.pos = n);
      Some record
    end

  method custom_get_path (row:custom_record) : Gtk.tree_path =
    GTree.Path.create [row.pos]

  method custom_get_value (row:custom_record) (column:int) (value:Gobject.g_value) : unit =
    Gobject.Value.init value column_types.(column);
    if row.pos >= num_rows then () 
    else if column = col_record then Gobject.Value.set value (`CAML (Obj.repr row))
    else if column = col_name then Gobject.Value.set value (`STRING (Some row.name))
    else if column = col_year_born then Gobject.Value.set value (`INT row.year_born)
    else ()

  method custom_iter_next (row:custom_record) : custom_record option =
    (* /* Is this the last record in the list? */ *)
    if row.pos + 1 >= num_rows then None else begin
      let nextrecord = rows.(row.pos + 1) in
      assert (nextrecord.pos = row.pos + 1);
      Some nextrecord
    end

  method custom_iter_children (rowopt:custom_record option) : custom_record option =
    begin match rowopt with
      | Some _ -> None (* /* this is a list, nodes have no children */ *)
      | None ->  (* /* parent == NULL is a special case; 
		       we need to return the first top-level row */ *) 
	  if num_rows = 0 then None 
	  else Some rows.(0)
    end

  method custom_iter_has_child (row:custom_record) : bool =
    false

  method custom_iter_n_children (rowopt:custom_record option) : int =
    begin match rowopt with
      | None -> num_rows (* /* special case: if iter == NULL, 
			       return number of top-level rows */ *)
      | _ -> 0 (* /* otherwise, this is easy again for a list */ *)
    end

  method custom_iter_nth_child (rowopt:custom_record option) (n:int) : custom_record option =
    begin match rowopt with
      | Some _ -> None (* /* a list has only top-level rows */ *)
      | None -> (* /* special case: if parent == NULL, 
		      set iter to n-th top-level row */ *)
	  if n >= num_rows then None 
	  else begin
	    let record = rows.(n) in
	    assert (record.pos = n);
	    Some record
	  end
    end

  method custom_iter_parent (row:custom_record) : custom_record option =
    None

  method append_record (name:string) (year_born:int) : unit =
    (* This function messes with the structure of the model
       but is called outside of GTK.  Therefore it must be
       synchronized with GTK's checking the model with
       Custom_model.sync or Custom_model.async *)
    Custom_model.async begin fun () ->
      let pos = num_rows in
      let newrecord = {
	name = name;
	(* name_collate_key = g_utf_collate_key name (-1); /* for fast sorting, used later */ *)
	year_born = year_born;
	pos = pos
      } 
      in
      num_rows <- num_rows + 1;
      rows <- 
        Array.init num_rows (fun n -> if n+1 = num_rows then newrecord else rows.(n));
      (* /* inform the tree view and other interested objects
         *  (e.g. tree row references) that we have inserted
         *  a new row, and where it was inserted */ *)
      self#custom_row_inserted (GTree.Path.create [pos]) newrecord
    end
end
    
(**********
 * main.c *
 **********)

let fill_model (custom_list:custom_list) : unit =
  let firstnames = ["Joe"; "Jane"; "William"; "Hannibal"; "Timothy"; "Gargamel"] in
  let surnames = ["Grokowich"; "Twitch"; "Borheimer"; "Bork"] in
  List.iter (fun sname ->
    List.iter (fun fname ->
      let name = fname ^ " " ^ sname in
      custom_list#append_record name (1900 + Random.int 103)) firstnames) surnames

let create_view_and_model () : GTree.view =
  let custom_list = new custom_list in
  fill_model custom_list;
  let view = GTree.view ~model:custom_list () in
  let renderer = GTree.cell_renderer_text [] in
  let col = GTree.view_column ~title:"Name" 
	      ~renderer:(renderer,["text",index_to_mock_column col_name]) () in
  ignore (view#append_column col);
  let renderer = GTree.cell_renderer_text [] in
  let col = GTree.view_column ~title:"Year Born" 
	      ~renderer:(renderer,["text",index_to_mock_column col_year_born]) () in
  ignore (view#append_column col);
  view

let _ =
  Random.self_init ();
  let gui_thread = Custom_model.start_gui () in
  let window = GWindow.window () in
  window#resize ~width:200 ~height:400;
  ignore (window#event#connect#delete ~callback:(fun _ -> GMain.Main.quit (); false));
  let scrollwin = GBin.scrolled_window ~packing:window#add () in
  let view = create_view_and_model () in
  scrollwin#add view#coerce;
  window#show ();
  Thread.join gui_thread
